@@ -44,24 +44,42 @@
 #include "gedit2.h"
 #include "gedit-menus.h"
 #include "gedit-debug.h"
-#include "gedit-prefs.h"
+#include "gedit-prefs-manager.h"
 #include "gedit-recent.h" 
 #include "gedit-file.h"
 #include "gedit-view.h"
 #include "gedit-utils.h"
 #include "gedit-plugins-engine.h"
-#include "gnome-recent-view-bonobo.h"
+#include "gedit-output-window.h"
+#include "recent-files/egg-recent-view-bonobo.h"
+#include "recent-files/egg-recent-view-gtk.h"
+#include "recent-files/egg-recent-model.h"
 
 #include <bonobo/bonobo-ui-util.h>
 #include <bonobo/bonobo-control.h>
 
 #include <gconf/gconf-client.h>
 
-#define RECENT_KEY "GeditRecent"
+#define RECENT_KEY 		"GeditRecent"
+#define OUTPUT_WINDOW_KEY	"GeditOutputWindow"
 
 struct _GeditMDIPrivate
 {
-	gint untitled_number;
+	gint dummy;
+};
+
+typedef struct _GeditWindowPrefs	GeditWindowPrefs;
+
+struct _GeditWindowPrefs
+{
+	gboolean toolbar_visible;
+	GeditToolbarSetting toolbar_buttons_style;
+
+	gboolean statusbar_visible;
+	gboolean statusbar_show_cursor_position;
+	gboolean statusbar_show_overwrite_mode;
+
+	gboolean output_window_visible;
 };
 
 static void gedit_mdi_class_init 	(GeditMDIClass	*klass);
@@ -87,6 +105,8 @@ static void gedit_mdi_child_state_changed_handler (GeditMDIChild *child);
 
 static void gedit_mdi_set_active_window_undo_redo_verbs_sensitivity (BonoboMDI *mdi);
 
+static void gedit_mdi_app_destroy_handler (BonoboMDI *mdi, BonoboWindow *window);
+
 static void gedit_mdi_view_menu_item_toggled_handler (
 			BonoboUIComponent           *ui_component,
 			const char                  *path,
@@ -94,8 +114,30 @@ static void gedit_mdi_view_menu_item_toggled_handler (
 			const char                  *state,
 			BonoboWindow                *win);
 
+
+static GQuark window_prefs_id = 0;
+
+static GeditWindowPrefs *gedit_window_prefs_new 		(void);
+static void		 gedit_window_prefs_attach_to_window 	(GeditWindowPrefs *prefs,
+								 BonoboWindow 	  *win);
+static GeditWindowPrefs	*gedit_window_prefs_get_from_window 	(BonoboWindow     *win);
+static void		 gedit_window_prefs_save 		(GeditWindowPrefs *prefs);
+
 static BonoboMDIClass *parent_class = NULL;
 
+enum
+{
+	TARGET_URI_LIST = 100
+};
+
+static GtkTargetEntry drag_types[] =
+{
+	{ "text/uri-list", 0, TARGET_URI_LIST },
+};
+
+static gint n_drag_types = sizeof (drag_types) / sizeof (drag_types [0]);
+
+
 GType
 gedit_mdi_get_type (void)
 {
@@ -135,20 +177,127 @@ gedit_mdi_class_init (GeditMDIClass *klass)
   	object_class->finalize = gedit_mdi_finalize;
 }
 
+static void
+menu_position_under_widget (GtkMenu *menu, int *x, int *y,
+			    gboolean *push_in, gpointer user_data)
+{
+	GtkWidget *w;
+	int screen_width, screen_height;
+	GtkRequisition requisition;
+
+	w = GTK_WIDGET (user_data);
+	
+	gdk_window_get_origin (w->window, x, y);
+	*x += w->allocation.x;
+	*y += w->allocation.y + w->allocation.height;
+
+	gtk_widget_size_request (GTK_WIDGET (menu), &requisition);
+
+	screen_width = gdk_screen_width ();
+	screen_height = gdk_screen_height ();
+
+	*x = CLAMP (*x, 0, MAX (0, screen_width - requisition.width));
+	*y = CLAMP (*y, 0, MAX (0, screen_height - requisition.height));
+}
+
+static gboolean
+open_button_pressed_cb (GtkWidget *widget,
+			      GdkEventButton *event,
+			      gpointer *user_data)
+{
+	GtkWidget *menu;
+	GeditMDI *mdi;
+
+	g_return_val_if_fail (GTK_IS_BUTTON (widget), FALSE);
+	g_return_val_if_fail (GEDIT_IS_MDI (user_data), FALSE);
+
+	mdi = GEDIT_MDI (user_data);
+
+	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget), TRUE);
+
+	menu = g_object_get_data (G_OBJECT (widget), "recent-menu");
+	gnome_popup_menu_do_popup_modal (menu,
+				menu_position_under_widget, widget,
+				event, widget, widget);
+	
+	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget), FALSE);
+	
+	return TRUE;
+}
+
+static gboolean
+open_button_key_pressed_cb (GtkWidget *widget,
+				  GdkEventKey *event,
+				  gpointer *user_data)
+{
+	if (event->keyval == GDK_space ||
+	    event->keyval == GDK_KP_Space ||
+	    event->keyval == GDK_Return ||
+	    event->keyval == GDK_KP_Enter) {
+		open_button_pressed_cb (widget, NULL, user_data);
+	}
+
+	return FALSE;
+}
+
+
+
+static void
+gedit_mdi_add_open_button (GeditMDI *mdi, BonoboUIComponent *ui_component,
+			 const gchar *path, const gchar *tooltip)
+{
+	GtkWidget *menu;
+	EggRecentViewGtk *view;
+	EggRecentModel *model;
+	GtkWidget *button;
+
+	button = gtk_toggle_button_new ();
+	gtk_button_set_relief (GTK_BUTTON (button), GTK_RELIEF_NONE);
+
+	gtk_container_add (GTK_CONTAINER (button),
+			   gtk_arrow_new (GTK_ARROW_DOWN, GTK_SHADOW_OUT));
+
+	gtk_widget_show_all (GTK_WIDGET (button));
+
+	model = gedit_recent_get_model ();
+
+	menu = gtk_menu_new ();
+	gtk_widget_show (menu);
+	view = egg_recent_view_gtk_new (menu, NULL);
+	g_signal_connect (view, "activate",
+			  G_CALLBACK (gedit_file_open_recent), NULL);
+	egg_recent_view_gtk_show_icons (view, TRUE);
+	egg_recent_view_gtk_show_numbers (view, FALSE);
+	egg_recent_view_set_model (EGG_RECENT_VIEW (view), model);
+	g_object_set_data (G_OBJECT (button), "recent-menu", menu);
+	
+	g_signal_connect_object (button, "key_press_event",
+				 G_CALLBACK (open_button_key_pressed_cb),
+				 mdi, 0);
+	g_signal_connect_object (button, "button_press_event",
+				 G_CALLBACK (open_button_pressed_cb),
+				 mdi, 0);
+
+	bonobo_ui_component_widget_set (ui_component, path, GTK_WIDGET (button),
+					NULL);
+}
+
 static void 
 gedit_mdi_init (GeditMDI  *mdi)
 {
 	gedit_debug (DEBUG_MDI, "START");
 
-	bonobo_mdi_construct (BONOBO_MDI (mdi), "gedit-2", "gedit");
+	bonobo_mdi_construct (BONOBO_MDI (mdi), 
+			      "gedit-2", 
+			      "gedit",
+			      gedit_prefs_manager_get_default_window_width (),
+			      gedit_prefs_manager_get_default_window_height ());
 	
 	mdi->priv = g_new0 (GeditMDIPrivate, 1);
 
-	mdi->priv->untitled_number = 0;	
-
 	bonobo_mdi_set_ui_template_file (BONOBO_MDI (mdi), GEDIT_UI_DIR "gedit-ui.xml", gedit_verbs);
 	
-	bonobo_mdi_set_child_list_path (BONOBO_MDI (mdi), "/menu/Documents/");
+	bonobo_mdi_set_child_list_path (BONOBO_MDI (mdi), "/menu/Documents/OpenDocuments/");
 
 	/* Connect signals */
 	g_signal_connect (G_OBJECT (mdi), "top_window_created",
@@ -168,12 +317,14 @@ gedit_mdi_init (GeditMDI  *mdi)
 			  G_CALLBACK (gedit_mdi_child_changed_handler), NULL);
 	g_signal_connect (G_OBJECT (mdi), "view_changed",
 			  G_CALLBACK (gedit_mdi_view_changed_handler), NULL);
-
 	
 	g_signal_connect (G_OBJECT (mdi), "all_windows_destroyed",
 			  G_CALLBACK (gedit_file_exit), NULL);
-			  
 
+	g_signal_connect (G_OBJECT (mdi), "top_window_destroy",
+			  G_CALLBACK (gedit_mdi_app_destroy_handler), NULL);
+
+			  
 	gedit_debug (DEBUG_MDI, "END");
 }
 
@@ -217,39 +368,6 @@ gedit_mdi_new (void)
 	return mdi;
 }
 
-/* Computes a unique ID string for use as the window role */
-/* Taken from EOG */
-
-static char *
-gen_role (void)
-{
-        char *ret;
-	static char *hostname;
-	time_t t;
-	static int serial;
-
-	t = time (NULL);
-
-	if (!hostname) {
-		static char buffer [512];
-
-		if ((gethostname (buffer, sizeof (buffer) - 1) == 0) &&
-		    (buffer [0] != 0))
-			hostname = buffer;
-		else
-			hostname = "localhost";
-	}
-
-	ret = g_strdup_printf ("gedit-window-%d-%d-%d-%ld-%d@%s",
-			       getpid (),
-			       getgid (),
-			       getppid (),
-			       (long) t,
-			       serial++,
-			       hostname);
-
-	return ret;
-}
 
 static void
 gedit_mdi_app_created_handler (BonoboMDI *mdi, BonoboWindow *win)
@@ -257,17 +375,11 @@ gedit_mdi_app_created_handler (BonoboMDI *mdi, BonoboWindow *win)
 	GtkWidget *widget;
 	BonoboControl *control;
 	BonoboUIComponent *ui_component;
-	GnomeRecentView *view;
-	GnomeRecentModel *model;
-	gchar *role;
-
-	static GtkTargetEntry drag_types[] =
-	{
-		{ "text/uri-list", 0, 0 },
-	};
-
-	static gint n_drag_types = sizeof (drag_types) / sizeof (drag_types [0]);
-
+	EggRecentView *view;
+	EggRecentModel *model;
+	GeditWindowPrefs *prefs;
+	GdkWindowState state;
+	
 	gedit_debug (DEBUG_MDI, "");
 	
 	ui_component = bonobo_mdi_get_ui_component_from_window (win);
@@ -316,11 +428,23 @@ gedit_mdi_app_created_handler (BonoboMDI *mdi, BonoboWindow *win)
 
 	g_object_set_data (G_OBJECT (win), "OverwriteMode", widget);
 
+	/* Add custom Open button to toolbar */
+	gedit_mdi_add_open_button (GEDIT_MDI (mdi),
+				   ui_component, "/Toolbar/FileOpenMenu",
+				   _("Open a file."));
+
+	prefs = gedit_window_prefs_new ();
+	gedit_window_prefs_attach_to_window (prefs, win);
+
 	/* Set the statusbar style according to prefs */
 	gedit_mdi_set_app_statusbar_style (win);
 	
 	/* Set the toolbar style according to prefs */
 	gedit_mdi_set_app_toolbar_style (win);
+
+	gedit_menus_set_verb_state (ui_component, 
+				    "/commands/ViewOutputWindow",
+				    prefs->output_window_visible);
 		
 	/* Add listener fo the view menu */
 	bonobo_ui_component_add_listener (ui_component, "ViewToolbar", 
@@ -329,6 +453,9 @@ gedit_mdi_app_created_handler (BonoboMDI *mdi, BonoboWindow *win)
 	bonobo_ui_component_add_listener (ui_component, "ViewStatusbar", 
 			(BonoboUIListenerFn)gedit_mdi_view_menu_item_toggled_handler, 
 			(gpointer)win);
+	bonobo_ui_component_add_listener (ui_component, "ViewOutputWindow", 
+			(BonoboUIListenerFn)gedit_mdi_view_menu_item_toggled_handler, 
+			(gpointer)win);
 
 	bonobo_ui_component_add_listener (ui_component, "ToolbarSystem", 
 			(BonoboUIListenerFn)gedit_mdi_view_menu_item_toggled_handler, 
@@ -339,6 +466,9 @@ gedit_mdi_app_created_handler (BonoboMDI *mdi, BonoboWindow *win)
 	bonobo_ui_component_add_listener (ui_component, "ToolbarIconText", 
 			(BonoboUIListenerFn)gedit_mdi_view_menu_item_toggled_handler, 
 			(gpointer)win);
+	bonobo_ui_component_add_listener (ui_component, "ToolbarIconBothHoriz", 
+			(BonoboUIListenerFn)gedit_mdi_view_menu_item_toggled_handler, 
+			(gpointer)win);
 
 	bonobo_ui_component_add_listener (ui_component, "StatusBarCursorPosition", 
 			(BonoboUIListenerFn)gedit_mdi_view_menu_item_toggled_handler, 
@@ -346,33 +476,62 @@ gedit_mdi_app_created_handler (BonoboMDI *mdi, BonoboWindow *win)
 	bonobo_ui_component_add_listener (ui_component, "StatusBarOverwriteMode", 
 			(BonoboUIListenerFn)gedit_mdi_view_menu_item_toggled_handler, 
 			(gpointer)win);
-
-
+	
 	/* add a GeditRecentView object */
 	model = gedit_recent_get_model ();
-	view = GNOME_RECENT_VIEW (gnome_recent_view_bonobo_new (ui_component, "/menu/File/Recents"));
-	gnome_recent_view_set_model (view, model);
+	view = EGG_RECENT_VIEW (egg_recent_view_bonobo_new (
+					ui_component, "/menu/File/Recents"));
+	egg_recent_view_bonobo_show_icons (EGG_RECENT_VIEW_BONOBO (view), FALSE);
+	egg_recent_view_set_model (view, model);
 	
 	g_signal_connect (G_OBJECT (view), "activate",
 			  G_CALLBACK (gedit_file_open_recent), NULL);
-	g_object_set_data (G_OBJECT (win), RECENT_KEY, view);
+	
+	g_object_set_data_full (G_OBJECT (win), RECENT_KEY, view,
+				g_object_unref);
 
-	/* Set the window prefs. */
-	gtk_window_set_default_size (GTK_WINDOW (win), 
-			gedit_settings->window_width, 
-			gedit_settings->window_height);
+		
+	/* Set window state and size, but only if the session is not being restored */
+	if (!bonobo_mdi_get_restoring_state (mdi))
+	{
+		state = gedit_prefs_manager_get_window_state ();
 
-	g_signal_connect (G_OBJECT (win), "configure_event",
-	                  G_CALLBACK (gedit_prefs_configure_event_handler), 
-			  NULL);
+		if ((state & GDK_WINDOW_STATE_MAXIMIZED) != 0)
+		{
+			gtk_window_set_default_size (GTK_WINDOW (win),
+						     gedit_prefs_manager_get_default_window_width (),
+						     gedit_prefs_manager_get_default_window_height ());
+
+			gtk_window_maximize (GTK_WINDOW (win));
+		}
+		else
+		{
+			gtk_window_set_default_size (GTK_WINDOW (win), 
+						     gedit_prefs_manager_get_window_width (),
+						     gedit_prefs_manager_get_window_height ());
+
+			gtk_window_unmaximize (GTK_WINDOW (win));
+		}
+
+		if ((state & GDK_WINDOW_STATE_STICKY ) != 0)
+			gtk_window_stick (GTK_WINDOW (win));
+		else
+			gtk_window_unstick (GTK_WINDOW (win));
+	}
 	
 	/* Add the plugins menus */
 	gedit_plugins_engine_update_plugins_ui (win, TRUE);
+}
 
-	/* Set window role */
-	role = gen_role ();
-	gtk_window_set_role (GTK_WINDOW (win), role);
-	g_free (role);
+static void
+gedit_mdi_app_destroy_handler (BonoboMDI *mdi, BonoboWindow *window)
+{
+	gedit_debug (DEBUG_MDI, "");
+	
+	g_return_if_fail (window != NULL);
+	g_return_if_fail (BONOBO_IS_WINDOW (window));
+
+	gedit_prefs_manager_save_window_size_and_state (window);
 }
 
 static void
@@ -384,67 +543,126 @@ gedit_mdi_view_menu_item_toggled_handler (
 			BonoboWindow                *win)
 {
 	gboolean s;
+	GeditWindowPrefs *prefs;
 
 	gedit_debug (DEBUG_MDI, "%s toggled to '%s'", path, state);
 
+	prefs = gedit_window_prefs_get_from_window (win);
+	g_return_if_fail (prefs != NULL);
+
 	s = (strcmp (state, "1") == 0);
 
-	if ((strcmp (path, "ViewToolbar") == 0) &&
-	    (s != gedit_settings->toolbar_visible))
+	if (strcmp (path, "ViewToolbar") == 0)
 	{
-		gedit_settings->toolbar_visible = s;
-		gedit_mdi_set_app_toolbar_style (win);
+		if (s != prefs->toolbar_visible)
+		{
+			prefs->toolbar_visible = s;
+			gedit_mdi_set_app_toolbar_style (win);
+		}
+
+		goto save_prefs;
+	}
+
+	if (strcmp (path, "ViewStatusbar") == 0)
+	{
+		if (s != prefs->statusbar_visible)
+		{
+			prefs->statusbar_visible = s;
+			gedit_mdi_set_app_statusbar_style (win);
+		}
+
+		goto save_prefs;
+	}
+
+	if (strcmp (path, "ViewOutputWindow") == 0)
+	{
+		if (s != prefs->output_window_visible)
+		{
+			GtkWidget *ow;
+
+			prefs->output_window_visible = s;
+			gedit_mdi_set_app_statusbar_style (win);
+
+			ow = gedit_mdi_get_output_window_from_window (win);
+
+			if (prefs->output_window_visible)
+				gtk_widget_show (ow);
+			else
+				gtk_widget_hide (ow);
+		}	
 
 		return;
 	}
 
-	if ((strcmp (path, "ViewStatusbar") == 0) &&
-	    (s != gedit_settings->statusbar_visible))
+
+	if (s && (strcmp (path, "ToolbarSystem") == 0))
 	{
-		gedit_settings->statusbar_visible = s;
-		gedit_mdi_set_app_statusbar_style (win);
+		if (prefs->toolbar_buttons_style  != GEDIT_TOOLBAR_SYSTEM)
+		{
+			prefs->toolbar_buttons_style  = GEDIT_TOOLBAR_SYSTEM;
+			gedit_mdi_set_app_toolbar_style (win);
+		}
+
+		goto save_prefs;
 	}
 
-	if (s && (strcmp (path, "ToolbarSystem") == 0) &&
-	    (gedit_settings->toolbar_buttons_style != GEDIT_TOOLBAR_SYSTEM))
-	{		
-		gedit_settings->toolbar_buttons_style = GEDIT_TOOLBAR_SYSTEM;
-		gedit_mdi_set_app_toolbar_style (win);
-
-		return;
-	}
-
-	if (s && (strcmp (path, "ToolbarIcon") == 0) &&
-	    (gedit_settings->toolbar_buttons_style != GEDIT_TOOLBAR_ICONS))
-	{		
-		gedit_settings->toolbar_buttons_style = GEDIT_TOOLBAR_ICONS;
-		gedit_mdi_set_app_toolbar_style (win);
-
-		return;
-	}
-
-	if (s && (strcmp (path, "ToolbarIconText") == 0) &&
-	    (gedit_settings->toolbar_buttons_style != GEDIT_TOOLBAR_ICONS_AND_TEXT))
-	{		
-		gedit_settings->toolbar_buttons_style = GEDIT_TOOLBAR_ICONS_AND_TEXT;
-		gedit_mdi_set_app_toolbar_style (win);
-
-		return;
-	}
-
-	if ((strcmp (path, "StatusBarCursorPosition") == 0) &&
-	    (s != gedit_settings->statusbar_view_cursor_position))
+	if (s && (strcmp (path, "ToolbarIcon") == 0))
 	{
-		gedit_settings->statusbar_view_cursor_position = s;
-		gedit_mdi_set_app_statusbar_style (win);
+		if (prefs->toolbar_buttons_style != GEDIT_TOOLBAR_ICONS)
+		{
+			prefs->toolbar_buttons_style = GEDIT_TOOLBAR_ICONS;
+			gedit_mdi_set_app_toolbar_style (win);
+		}
+		
+		goto save_prefs;
 	}
 
-	if ((strcmp (path, "StatusBarOverwriteMode") == 0) &&
-	    (s != gedit_settings->statusbar_view_overwrite_mode))
+	if (s && (strcmp (path, "ToolbarIconText") == 0))
 	{
-		gedit_settings->statusbar_view_overwrite_mode = s;
-		gedit_mdi_set_app_statusbar_style (win);
+		if (prefs->toolbar_buttons_style != GEDIT_TOOLBAR_ICONS_AND_TEXT)
+		{
+			prefs->toolbar_buttons_style = GEDIT_TOOLBAR_ICONS_AND_TEXT;
+			gedit_mdi_set_app_toolbar_style (win);
+		}
+
+		goto save_prefs;
 	}
+
+	if (s && (strcmp (path, "ToolbarIconBothHoriz") == 0))
+	{
+		if (prefs->toolbar_buttons_style != GEDIT_TOOLBAR_ICONS_BOTH_HORIZ)
+		{
+			prefs->toolbar_buttons_style = GEDIT_TOOLBAR_ICONS_BOTH_HORIZ;
+			gedit_mdi_set_app_toolbar_style (win);
+		}
+
+		goto save_prefs;
+	}
+
+	if (strcmp (path, "StatusBarCursorPosition") == 0)
+	{
+		if (s != prefs->statusbar_show_cursor_position)
+		{
+			prefs->statusbar_show_cursor_position = s;
+			gedit_mdi_set_app_statusbar_style (win);
+		}
+
+		goto save_prefs;
+	}
+
+	if (strcmp (path, "StatusBarOverwriteMode") == 0)
+	{
+		if (s != prefs->statusbar_show_overwrite_mode)
+		{
+			prefs->statusbar_show_overwrite_mode = s;
+			gedit_mdi_set_app_statusbar_style (win);
+		}
+
+		goto save_prefs;
+	}
+
+save_prefs:
+	gedit_window_prefs_save (prefs);
 }
 
 static void 
@@ -458,6 +676,9 @@ gedit_mdi_drag_data_received_handler (GtkWidget *widget, GdkDragContext *context
 	
 	gedit_debug (DEBUG_MDI, "");
 
+	if (info != TARGET_URI_LIST)
+		return;
+			
 	list = gnome_vfs_uri_list_parse (selection_data->data);
 	p = list;
 
@@ -471,7 +692,7 @@ gedit_mdi_drag_data_received_handler (GtkWidget *widget, GdkDragContext *context
 	
 	gnome_vfs_uri_list_free (list);
 
-	gedit_file_open_uri_list (file_list, 0);	
+	gedit_file_open_uri_list (file_list, 0, FALSE);	
 	
 	if (file_list == NULL)
 		return;
@@ -487,11 +708,15 @@ static void
 gedit_mdi_set_app_toolbar_style (BonoboWindow *win)
 {
 	BonoboUIComponent *ui_component;
+	GeditWindowPrefs *prefs = NULL;
 
 	gedit_debug (DEBUG_MDI, "");
 	
 	g_return_if_fail (BONOBO_IS_WINDOW (win));
-			
+	
+	prefs = gedit_window_prefs_get_from_window (win);
+	g_return_if_fail (prefs != NULL);
+	
 	ui_component = bonobo_mdi_get_ui_component_from_window (win);
 	g_return_if_fail (ui_component != NULL);
 			
@@ -500,34 +725,43 @@ gedit_mdi_set_app_toolbar_style (BonoboWindow *win)
 	/* Updated view menu */
 	gedit_menus_set_verb_state (ui_component, 
 				    "/commands/ViewToolbar",
-				    gedit_settings->toolbar_visible);
+				    prefs->toolbar_visible);
 
 	gedit_menus_set_verb_sensitive (ui_component, 
 				        "/commands/ToolbarSystem",
-				        gedit_settings->toolbar_visible);
+				        prefs->toolbar_visible);
 	gedit_menus_set_verb_sensitive (ui_component, 
 				        "/commands/ToolbarIcon",
-				        gedit_settings->toolbar_visible);
+				        prefs->toolbar_visible);
 	gedit_menus_set_verb_sensitive (ui_component, 
 				        "/commands/ToolbarIconText",
-				        gedit_settings->toolbar_visible);
+				        prefs->toolbar_visible);
+	gedit_menus_set_verb_sensitive (ui_component, 
+				        "/commands/ToolbarIconBothHoriz",
+				        prefs->toolbar_visible);
+	gedit_menus_set_verb_sensitive (ui_component, 
+				        "/commands/ToolbarTooltips",
+				        prefs->toolbar_visible);
 
 	gedit_menus_set_verb_state (ui_component, 
 				    "/commands/ToolbarSystem",
-				    gedit_settings->toolbar_buttons_style == GEDIT_TOOLBAR_SYSTEM);
+				    prefs->toolbar_buttons_style == GEDIT_TOOLBAR_SYSTEM);
 
 	gedit_menus_set_verb_state (ui_component, 
 				    "/commands/ToolbarIcon",
-				    gedit_settings->toolbar_buttons_style == GEDIT_TOOLBAR_ICONS);
+				    prefs->toolbar_buttons_style == GEDIT_TOOLBAR_ICONS);
 
 	gedit_menus_set_verb_state (ui_component, 
 				    "/commands/ToolbarIconText",
-				    gedit_settings->toolbar_buttons_style == GEDIT_TOOLBAR_ICONS_AND_TEXT);
+				    prefs->toolbar_buttons_style == GEDIT_TOOLBAR_ICONS_AND_TEXT);
 
-	switch (gedit_settings->toolbar_buttons_style)
+	gedit_menus_set_verb_state (ui_component, 
+				    "/commands/ToolbarIconBothHoriz",
+				    prefs->toolbar_buttons_style == GEDIT_TOOLBAR_ICONS_BOTH_HORIZ);
+	
+	switch (prefs->toolbar_buttons_style)
 	{
 		case GEDIT_TOOLBAR_SYSTEM:
-
 			gedit_debug (DEBUG_MDI, "GEDIT: SYSTEM");
 			bonobo_ui_component_set_prop (
 				ui_component, "/Toolbar", "look", "system", NULL);
@@ -547,6 +781,13 @@ gedit_mdi_set_app_toolbar_style (BonoboWindow *win)
 				ui_component, "/Toolbar", "look", "both", NULL);
 			
 			break;
+			
+		case GEDIT_TOOLBAR_ICONS_BOTH_HORIZ:
+			gedit_debug (DEBUG_MDI, "GEDIT: ICONS_BOTH_HORIZ");
+			bonobo_ui_component_set_prop (
+				ui_component, "/Toolbar", "look", "both_horiz", NULL);
+			
+			break;       
 		default:
 			goto error;
 			break;
@@ -554,7 +795,7 @@ gedit_mdi_set_app_toolbar_style (BonoboWindow *win)
 	
 	bonobo_ui_component_set_prop (
 			ui_component, "/Toolbar",
-			"hidden", gedit_settings->toolbar_visible ? "0":"1", NULL);
+			"hidden", prefs->toolbar_visible ? "0":"1", NULL);
 
  error:
 	bonobo_ui_component_thaw (ui_component, NULL);
@@ -563,51 +804,54 @@ gedit_mdi_set_app_toolbar_style (BonoboWindow *win)
 static void
 gedit_mdi_set_app_statusbar_style (BonoboWindow *win)
 {
+	GeditWindowPrefs *prefs = NULL;
 	BonoboUIComponent *ui_component;
 	GtkWidget *cp, *om;
 	
 	gedit_debug (DEBUG_MDI, "");
 	
 	g_return_if_fail (BONOBO_IS_WINDOW (win));
-			
+
+	prefs = gedit_window_prefs_get_from_window (win);
+	g_return_if_fail (prefs != NULL);
+
 	ui_component = bonobo_mdi_get_ui_component_from_window (win);
 	g_return_if_fail (ui_component != NULL);
 
 	bonobo_ui_component_freeze (ui_component, NULL);
-
+	
 	/* Update menu */
 	gedit_menus_set_verb_state (ui_component, 
 				    "/commands/ViewStatusbar",
-				    gedit_settings->statusbar_visible);
+				    prefs->statusbar_visible);
 
 	gedit_menus_set_verb_sensitive (ui_component, 
 				    "/commands/StatusBarCursorPosition",
-				    gedit_settings->statusbar_visible);
+				    prefs->statusbar_visible);
 
 	gedit_menus_set_verb_sensitive (ui_component, 
 				    "/commands/StatusBarOverwriteMode",
-				    gedit_settings->statusbar_visible);
+				    prefs->statusbar_visible);
 
 	gedit_menus_set_verb_state (ui_component, 
 				    "/commands/StatusBarCursorPosition",
-				    gedit_settings->statusbar_view_cursor_position);
+				    prefs->statusbar_show_cursor_position);
 
 	gedit_menus_set_verb_state (ui_component, 
 				    "/commands/StatusBarOverwriteMode",
-				    gedit_settings->statusbar_view_overwrite_mode);
-
+				    prefs->statusbar_show_overwrite_mode);
 	
 	/* Actually update status bar style */
 	bonobo_ui_component_set_prop (
 		ui_component, "/status",
-		"hidden", gedit_settings->statusbar_visible ? "0" : "1",
+		"hidden", prefs->statusbar_visible ? "0" : "1",
 		NULL);
 
 	cp = GTK_WIDGET (g_object_get_data (G_OBJECT (win), "CursorPosition"));
 	if (cp == NULL)
 		goto error;
 
-	if (gedit_settings->statusbar_view_cursor_position)
+	if (prefs->statusbar_show_cursor_position)
 	{
 		bonobo_ui_component_set_prop (
 			ui_component, "/status/CursorPosition", "hidden", "0", NULL);
@@ -626,7 +870,7 @@ gedit_mdi_set_app_statusbar_style (BonoboWindow *win)
 	if (om == NULL)
 		goto error;
 
-	if (gedit_settings->statusbar_view_overwrite_mode)
+	if (prefs->statusbar_show_overwrite_mode)
 	{
 		gtk_widget_show (om);
 		
@@ -641,13 +885,13 @@ gedit_mdi_set_app_statusbar_style (BonoboWindow *win)
 		gtk_widget_hide (om);
 	}
 
-	if (!gedit_settings->statusbar_view_overwrite_mode)
+	if (!prefs->statusbar_show_overwrite_mode)
 		gtk_statusbar_set_has_resize_grip (GTK_STATUSBAR (cp), TRUE);
 	else
 		gtk_statusbar_set_has_resize_grip (GTK_STATUSBAR (cp), FALSE);
 
-	if (!gedit_settings->statusbar_view_cursor_position &&
-	    !gedit_settings->statusbar_view_overwrite_mode)
+	if (!prefs->statusbar_show_cursor_position &&
+	    !prefs->statusbar_show_overwrite_mode)
 		bonobo_ui_component_set_prop (
 			ui_component, "/status", "resize_grip", "1", NULL);
 	else
@@ -700,25 +944,23 @@ gedit_mdi_add_child_handler (BonoboMDI *mdi, BonoboMDIChild *child)
 static gint 
 gedit_mdi_add_view_handler (BonoboMDI *mdi, GtkWidget *view)
 {
-	static GtkTargetEntry drag_types[] =
-	{
-		{ "text/uri-list", 0, 0 },
-	};
-
-	static gint n_drag_types = sizeof (drag_types) / sizeof (drag_types [0]);
+	GtkTextView *text_view;
+	GtkTargetList *tl;
 
 	gedit_debug (DEBUG_MDI, "");
+
+	g_return_val_if_fail (view != NULL, TRUE);
+
+	text_view = gedit_view_get_gtk_text_view (GEDIT_VIEW (view));
+	g_return_val_if_fail (text_view != NULL, TRUE);
 	
-	/* FIXME */
 	/* Drag and drop support */
-	gtk_drag_dest_set (view,
-			   GTK_DEST_DEFAULT_MOTION |
-			   GTK_DEST_DEFAULT_HIGHLIGHT |
-			   GTK_DEST_DEFAULT_DROP,
-			   drag_types, n_drag_types,
-			   GDK_ACTION_COPY);
-		
-	g_signal_connect (G_OBJECT (view), "drag_data_received",
+	tl = gtk_drag_dest_get_target_list (GTK_WIDGET (text_view));
+	g_return_val_if_fail (tl != NULL, TRUE);
+
+	gtk_target_list_add_table (tl, drag_types, n_drag_types);
+
+	g_signal_connect (G_OBJECT (text_view), "drag_data_received",
 			  G_CALLBACK (gedit_mdi_drag_data_received_handler), 
 			  NULL);
 
@@ -730,6 +972,8 @@ gedit_mdi_remove_child_handler (BonoboMDI *mdi, BonoboMDIChild *child)
 {
 	GeditDocument* doc;
 	gboolean close = TRUE;
+	gchar *raw_uri;
+	gboolean deleted = FALSE;
 	
 	gedit_debug (DEBUG_MDI, "");
 
@@ -738,7 +982,17 @@ gedit_mdi_remove_child_handler (BonoboMDI *mdi, BonoboMDIChild *child)
 
 	doc = GEDIT_MDI_CHILD (child)->document;
 
-	if (gedit_document_get_modified (doc))
+	raw_uri = gedit_document_get_raw_uri (doc); 
+	if (raw_uri != NULL)
+	{
+		if (gedit_document_is_readonly (doc))
+			deleted = FALSE;
+		else
+			deleted = !gedit_utils_uri_exists (raw_uri);
+	}
+	g_free (raw_uri);
+				
+	if (gedit_document_get_modified (doc) || deleted)
 	{
 		GtkWidget *msgbox, *w;
 		gchar *fname = NULL, *msg = NULL;
@@ -748,7 +1002,14 @@ gedit_mdi_remove_child_handler (BonoboMDI *mdi, BonoboMDIChild *child)
 		w = GTK_WIDGET (g_list_nth_data (bonobo_mdi_child_get_views (child), 0));
 			
 		if(w != NULL)
+		{
+			GtkWindow *window;
+
+			window = GTK_WINDOW (bonobo_mdi_get_window_from_view (w));
+			gtk_window_present (window);
+			
 			bonobo_mdi_set_active_view (mdi, w);
+		}
 
 		fname = gedit_document_get_short_name (doc);
 
@@ -798,6 +1059,9 @@ gedit_mdi_remove_child_handler (BonoboMDI *mdi, BonoboMDIChild *child)
 
 		gtk_window_set_resizable (GTK_WINDOW (msgbox), FALSE);
 
+		gtk_widget_show (msgbox);
+		gtk_window_present (GTK_WINDOW (msgbox));
+
 		ret = gtk_dialog_run (GTK_DIALOG (msgbox));
 		
 		gtk_widget_destroy (msgbox);
@@ -808,7 +1072,7 @@ gedit_mdi_remove_child_handler (BonoboMDI *mdi, BonoboMDIChild *child)
 		switch (ret)
 		{
 			case GTK_RESPONSE_YES:
-				close = gedit_file_save (GEDIT_MDI_CHILD (child));
+				close = gedit_file_save (GEDIT_MDI_CHILD (child), TRUE);
 				break;
 			case GTK_RESPONSE_NO:
 				close = TRUE;
@@ -986,7 +1250,7 @@ gedit_mdi_set_active_window_verbs_sensitivity (BonoboMDI *mdi)
 	bonobo_ui_component_freeze (ui_component, NULL);
 	
 	gedit_plugins_engine_update_plugins_ui (active_window, FALSE);
-
+	
 	if (active_child == NULL)
 	{
 		gedit_menus_set_verb_list_sensitive (ui_component, 
@@ -999,6 +1263,9 @@ gedit_mdi_set_active_window_verbs_sensitivity (BonoboMDI *mdi)
 				gedit_menus_all_sensible_verbs, TRUE);
 	}
 
+	gedit_menus_set_verb_sensitive (ui_component, "/commands/DocumentsMoveToNewWindow",
+				(bonobo_mdi_n_children_for_window (active_window) > 1) ? TRUE : FALSE);
+
 	doc = GEDIT_MDI_CHILD (active_child)->document;
 	g_return_if_fail (doc != NULL);
 	
@@ -1064,61 +1331,189 @@ gedit_mdi_set_active_window_undo_redo_verbs_sensitivity (BonoboMDI *mdi)
 	bonobo_ui_component_thaw (ui_component, NULL);
 }
 
-void
-gedit_mdi_update_ui_according_to_preferences (GeditMDI *mdi)
-{
-	GList *children;
-	GdkColor background;
-	GdkColor text;
-	GdkColor selection;
-	GdkColor sel_text;
-
-	gchar* font = NULL;
-
-	gedit_debug (DEBUG_MDI, "");
-
-	children = bonobo_mdi_get_children (BONOBO_MDI (mdi));
-
-	if (!gedit_settings->use_default_colors)
-	{
-		background = gedit_settings->background_color;
-		text = gedit_settings->text_color;
-		selection = gedit_settings->selection_color;
-		sel_text = gedit_settings->selected_text_color;
-	}
-
-	if (!gedit_settings->use_default_font)
-		font = g_strdup (gedit_settings->editor_font);
-
-	while (children != NULL)
-	{
-		GList *views = bonobo_mdi_child_get_views (BONOBO_MDI_CHILD (children->data));
-
-		while (views != NULL)
-		{
-			GeditView *v =	GEDIT_VIEW (views->data);
-			
-			gedit_view_set_font (v, gedit_settings->use_default_font, font);
-			gedit_view_set_colors (v, gedit_settings->use_default_colors, &background, &text, &selection, &sel_text);
-			gedit_view_set_wrap_mode (v, gedit_settings->wrap_mode);
-			gedit_view_show_line_numbers (v, gedit_settings->show_line_numbers);
-			gedit_view_set_tab_size (v, gedit_settings->tab_size);
-			views = views->next;
-		}
-		
-		children = children->next;
-	}
-
-	g_free (font);
-}
-
-GnomeRecentView *
-gedit_mdi_get_recent_view_from_window (BonoboWindow* win)
+EggRecentView *
+gedit_mdi_get_recent_view_from_window (BonoboWindow *win)
 {
 	gpointer r;
 	gedit_debug (DEBUG_MDI, "");
 
 	r = g_object_get_data (G_OBJECT (win), RECENT_KEY);
 	
-	return (r != NULL) ? GNOME_RECENT_VIEW (r) : NULL;
+	return (r != NULL) ? EGG_RECENT_VIEW (r) : NULL;
+}
+
+static void
+gedit_mdi_close_output_window_cb (GtkWidget *widget, gpointer user_data)
+{
+	BonoboWindow *bw;
+	
+	bw = BONOBO_WINDOW (user_data);
+
+	gtk_widget_hide (widget);
+}
+
+static void
+gedit_mdi_show_output_window_cb (GtkWidget *widget, gpointer user_data)
+{
+	BonoboWindow *bw;
+	GeditWindowPrefs *prefs;
+	BonoboUIComponent *ui_component;
+	
+	gedit_debug (DEBUG_MDI, "");
+	
+	bw = BONOBO_WINDOW (user_data);
+	g_return_if_fail (bw != NULL);
+	
+	ui_component = bonobo_mdi_get_ui_component_from_window (bw);
+	g_return_if_fail (ui_component != NULL);
+	
+	prefs = gedit_window_prefs_get_from_window (bw);
+	g_return_if_fail (prefs != NULL);
+	
+	prefs->output_window_visible = TRUE;
+	
+	gedit_menus_set_verb_state (ui_component, 
+				    "/commands/ViewOutputWindow",
+				    prefs->output_window_visible);
+}
+
+static void
+gedit_mdi_hide_output_window_cb (GtkWidget *widget, gpointer user_data)
+{
+	BonoboWindow *bw;
+	GeditWindowPrefs *prefs;
+	BonoboUIComponent *ui_component;
+	
+	gedit_debug (DEBUG_MDI, "");
+	
+	bw = BONOBO_WINDOW (user_data);
+	g_return_if_fail (bw != NULL);
+	
+	ui_component = bonobo_mdi_get_ui_component_from_window (bw);
+	g_return_if_fail (ui_component != NULL);
+
+	prefs = gedit_window_prefs_get_from_window (bw);
+	g_return_if_fail (prefs != NULL);
+
+	prefs->output_window_visible = FALSE;
+
+	gedit_menus_set_verb_state (ui_component, 
+				    "/commands/ViewOutputWindow",
+				    prefs->output_window_visible);
+}
+
+
+GtkWidget *
+gedit_mdi_get_output_window_from_window (BonoboWindow *win)
+{
+	gpointer r;
+	gedit_debug (DEBUG_MDI, "");
+
+	r = g_object_get_data (G_OBJECT (win), OUTPUT_WINDOW_KEY);
+
+	if (r == NULL)
+	{
+		GtkWidget *ow;
+
+		/* Add output window */
+		ow = gedit_output_window_new ();
+		bonobo_mdi_set_bottom_pane_for_window (BONOBO_WINDOW (win), ow);
+		gtk_widget_show_all (ow);
+		g_signal_connect (G_OBJECT (ow), "close_requested",
+			  G_CALLBACK (gedit_mdi_close_output_window_cb), win);
+		g_signal_connect (G_OBJECT (ow), "show",
+			  G_CALLBACK (gedit_mdi_show_output_window_cb), win);
+		g_signal_connect (G_OBJECT (ow), "hide",
+			  G_CALLBACK (gedit_mdi_hide_output_window_cb), win);
+		g_object_set_data (G_OBJECT (win), OUTPUT_WINDOW_KEY, ow);
+
+		return ow;
+	}
+	
+	return (r != NULL) ? GTK_WIDGET (r) : NULL;
+}
+
+
+
+static GeditWindowPrefs *
+gedit_window_prefs_new (void)
+{
+	GeditWindowPrefs *prefs;
+
+	gedit_debug (DEBUG_MDI, "");
+
+	prefs = g_new0 (GeditWindowPrefs, 1);
+
+	prefs->toolbar_visible = gedit_prefs_manager_get_toolbar_visible ();
+	prefs->toolbar_buttons_style = gedit_prefs_manager_get_toolbar_buttons_style ();
+
+	prefs->statusbar_visible = gedit_prefs_manager_get_statusbar_visible ();
+	prefs->statusbar_show_cursor_position = gedit_prefs_manager_get_statusbar_show_cursor_position ();
+	prefs->statusbar_show_overwrite_mode = gedit_prefs_manager_get_statusbar_show_overwrite_mode ();
+
+	prefs->output_window_visible = FALSE;
+
+	return prefs;
+}
+
+static void
+gedit_window_prefs_attach_to_window (GeditWindowPrefs *prefs, BonoboWindow *win)
+{
+	gedit_debug (DEBUG_MDI, "");
+
+	g_return_if_fail (prefs != NULL);
+	g_return_if_fail (win != NULL);
+	g_return_if_fail (BONOBO_IS_WINDOW (win));
+
+	if (!window_prefs_id)
+		window_prefs_id = g_quark_from_static_string ("GeditWindowPrefsData");
+
+	g_object_set_qdata_full (G_OBJECT (win), window_prefs_id, prefs, g_free);
+}
+
+static GeditWindowPrefs	*
+gedit_window_prefs_get_from_window (BonoboWindow *win)
+{
+	GeditWindowPrefs *prefs;
+
+	gedit_debug (DEBUG_MDI, "");
+
+	g_return_val_if_fail (win != NULL, NULL);
+	g_return_val_if_fail (BONOBO_IS_WINDOW (win), NULL);
+
+	prefs = g_object_get_qdata (G_OBJECT (win), window_prefs_id);
+
+	return (prefs != NULL) ? (GeditWindowPrefs*)prefs : NULL;
+}
+
+static void
+gedit_window_prefs_save (GeditWindowPrefs *prefs)
+{
+	gedit_debug (DEBUG_MDI, "");
+
+	g_return_if_fail (prefs != NULL);
+
+	if ((prefs->toolbar_visible != gedit_prefs_manager_get_toolbar_visible ()) &&
+	     gedit_prefs_manager_toolbar_visible_can_set ())
+		gedit_prefs_manager_set_toolbar_visible (prefs->toolbar_visible);
+
+	if ((prefs->toolbar_buttons_style != gedit_prefs_manager_get_toolbar_buttons_style ()) &&
+	    gedit_prefs_manager_toolbar_buttons_style_can_set ())
+		gedit_prefs_manager_set_toolbar_buttons_style (prefs->toolbar_buttons_style);
+
+	if ((prefs->statusbar_visible != gedit_prefs_manager_get_statusbar_visible ()) &&
+	    gedit_prefs_manager_get_statusbar_visible ())
+		gedit_prefs_manager_set_statusbar_visible (prefs->statusbar_visible);
+
+	if ((prefs->statusbar_show_cursor_position != 
+			gedit_prefs_manager_get_statusbar_show_cursor_position ()) &&
+	    gedit_prefs_manager_statusbar_show_cursor_position_can_set ())
+		gedit_prefs_manager_set_statusbar_show_cursor_position (
+				prefs->statusbar_show_cursor_position);
+
+	if ((prefs->statusbar_show_overwrite_mode != 
+			gedit_prefs_manager_get_statusbar_show_overwrite_mode ()) &&
+	    gedit_prefs_manager_statusbar_show_overwrite_mode_can_set ())
+		gedit_prefs_manager_set_statusbar_show_overwrite_mode (
+				prefs->statusbar_show_overwrite_mode);
 }
